//***************************************************************************
// Copyright 2007-2014 Universidade do Porto - Faculdade de Engenharia      *
// Laboratório de Sistemas e Tecnologia Subaquática (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// https://www.lsts.pt/dune/licence.                                        *
//***************************************************************************
// Author: Joel Cardoso                                                     *
// Author: Eduardo Marques                                                  *
// Author: Ricardo Martins                                                  *
// Author: Joao Fortuna                                                     *
// Author: Thor Audun Steen                                                 *
//***************************************************************************

// ISO C++ 98 headers.
#include <vector>
#include <cmath>

// DUNE headers.
#include <DUNE/DUNE.hpp>

// MAVLink headers.
#include <mavlink/ardupilotmega/mavlink.h>

namespace Control
{
  namespace UAV
  {
    namespace ArduCopter
    {
      using DUNE_NAMESPACES;

      //! APM Type specifier
      enum APM_Vehicle {
        //! Unset or unknown vehicle type
        VEHICLE_UNKNOWN = 0,
        //! Fixed wing types
        VEHICLE_FIXEDWING,
        //! Copter types (quad, hexa, etc)
        VEHICLE_COPTER
      };

      //! List of cloops
      enum APM_c_loops {
        ACL_ROLL,
        ACL_PITCH,
        ACL_YAW,
        ACL_VELOCITY,
        ACL_ALTITUDE,
        ACL_VERTICAL_RATE,
        ACL_CONTROL,
        ACL_MAX
      };

      //! List of arducopter modes
      enum APM_copterModes {
        CP_MODE_STABILIZE=0,                     // hold level position
        CP_MODE_ACRO=1,                          // rate control
        CP_MODE_ALT_HOLD=2,                      // AUTO control
        CP_MODE_AUTO=3,                          // AUTO control
        CP_MODE_GUIDED=4,                        // AUTO control
        CP_MODE_LOITER=5,                        // Hold a single location
        CP_MODE_RTL=6,                           // AUTO control
        CP_MODE_CIRCLE=7,                        // AUTO control
        CP_MODE_POSITION=8,                      // AUTO control
        CP_MODE_LAND=9,                          // AUTO control
        CP_MODE_OF_LOITER=10,                    // Hold a single location using optical flow sensor
        CP_MODE_DRIFT=11,                        // DRIFT mode (Note: 12 is no longer used)
        CP_MODE_DUNE=12,
        CP_MODE_SPORT=13                       // earth frame rate control
      };


      /*
       * Target system is uint8_t
       * Split into four fields a two bits
       * 0b 7 6 5 4 3 2 1 0
       *
       * Bits   Description
       * 0-1      How to use xy-component
       * 2-4    How to use z -component
       * 5      How to use roll-pitch component
       * 6-7    How to use yaw-component
       *
       */

      enum APM_target_masks {
        TARGET_MASK_XY   = 0b00000011,
        TARGET_MASK_Z    = 0b00011100,
        TARGET_MASK_RP   = 0b00100000,
        TARGET_MASK_YAW  = 0b11000000
      };

      enum APM_target_xy {
        // MASK:              0b00000011
        TARGET_XY_NONE      = 0b00000000,
        TARGET_XY_VEL_BODY  = 0b00000001,
        TARGET_XY_VEL_NED   = 0b00000010,
      };

      enum APM_target_z {
        // MASK:          0b00011100
        TARGET_Z_NONE   = 0b00000000,
        TARGET_Z_VEL    = 0b00000100,
        TARGET_Z_ALT    = 0b00001000,
        TARGET_Z_THRUST = 0b00001100,
        TARGET_Z_RC     = 0b00010000
      };

      enum APM_target_rp {
        // MASK:          0b00100000
        TARGET_RP_NONE  = 0b00000000,
        TARGET_RP_SET   = 0b00100000,
      };

      enum APM_target_yaw {
        // MASK:              0b11000000
        TARGET_YAW_NONE     = 0b00000000,
        TARGET_YAW_SET      = 0b01000000,
        TARGET_YAW_RC       = 0b11000000,
        TARGET_YAW_RESERVED = 0b10000000

      };


      //! %Task arguments.
      struct Arguments
      {
        //! Communications timeout
        uint8_t comm_timeout;
        //! Use Ardupilot's waypoint tracker
        bool ardu_tracker;
        //! Use ardupilot velocity control
        bool ardu_vel_ctrl;
        //! Use ardupilot body velocity control
        bool ardu_vel_ctrl_body;
        //! Dispatch customManeuver on DUNE-mode
        bool auto_custom_maneuver;
        //! String of custom maneuver
        std::string auto_custom_maneuver_name;
        std::string auto_custom_maneuver_plan_name;
        std::string auto_custom_maneuver_parameter;
        //! TCP Port
        uint16_t TCP_port;
        //! TCP Address
        Address TCP_addr;
        //! Telemetry Rate
        uint8_t trate;
        //! Default Altitude
        float alt;
        //! Default Speed
        float speed;
        //! GPS is uBlox
        bool ublox;
        //! LoiterHere (default) radius
        float lradius;
        //! Loitering tolerance
        int ltolerance;
        //! Has Power Module
        bool pwrm;
        //! WP seconds before reach
        int secs;
        //! WP meters before reach
        int wptolerance;
        //! Dune mode minimum interval between each new desired message for apm dune mode
        float min_command_interval;
        //! Which wp switch method to use
        bool wptmethod;
      };

      struct Task: public DUNE::Tasks::Task
      {
        //! Task arguments.
        Arguments m_args;
        //! Arduino packet handling
        typedef void (Task::* PktHandler)(const mavlink_message_t* msg);
        typedef std::map<int, PktHandler> PktHandlerMap;
        PktHandlerMap m_mlh;
        double m_last_pkt_time;
        uint8_t m_buf[512];
        //! Estimated state message.
        IMC::EstimatedState m_estate;
        //! Battery messages
        IMC::Voltage m_volt;
        IMC::Current m_curr;
        IMC::FuelLevel m_fuel;
        //! Pressure message
        IMC::Pressure m_pressure;
        //! Temperature message
        IMC::Temperature m_temp;
        //! GPS Fix message
        IMC::GpsFix m_fix;
        //! Wind message
        IMC::EstimatedStreamVelocity m_stream;
        //! IMU messages
        IMC::AngularVelocity m_ang_vel;
        IMC::Acceleration m_accel;
        IMC::MagneticField m_mag_field;
        //! Servo PWM message
        IMC::SetServoPosition m_servo;
        //! Path Control State
        IMC::PathControlState m_pcs;
        //! Reference Lat and Lon and Hei to measure displacement
        fp64_t ref_lat, ref_lon;
        fp32_t ref_hei;
        //! TCP socket
        TCPSocket* m_TCP_sock;
        Address m_TCP_addr;
        uint16_t m_TCP_port;
        //! System ID
        uint8_t m_sysid;
        //! Last received position
        double m_lat, m_lon;
        float m_alt;
        //! External control
        bool m_external;
        //! Current waypoint
        int m_current_wp;
        //! Critical WP
        bool m_critical;
        //! Bitfield of enabled control loops.
        uint32_t m_cloops;
        //! Parser Variables
        mavlink_message_t m_msg;
        int m_desired_radius;
        int m_gnd_speed;
        int m_mode;
        bool m_changing_wp;
        bool m_error_missing, m_error_ext;
        //! Setup rate hack
        bool m_has_setup_rate;
        //! Type of system to be controlled
        APM_Vehicle m_vehicle_type;
        // External control
        //IMC::DesiredExtCtrl m_extCtrl;
        // Prev desired vel
        IMC::DesiredVelocity m_d_velocity;
        // Desired Yaw
        IMC::DesiredYaw m_d_yaw;
        // Desired Altitude
        IMC::DesiredZ m_d_z;
        // Other desired stuff
        IMC::DesiredRoll m_d_roll;
        IMC::DesiredPitch m_d_pitch;
        IMC::DesiredControl m_d_control;
        //IMC::DesiredVerticalRate m_d_vertical_rate;
        // Use ardupilot tracker
        bool m_ardu_tracker;
        // Use ardupilot velocity control
        bool m_ardu_vel_ctrl;
        // Use ardupilot body velocity control
        bool m_ardu_vel_ctrl_body;
        // Auto-dispatch custom maneuver
        bool m_auto_custom_maneuver;
        // List of updated desired states,
        bool m_updated_desired[ACL_MAX];
        // Previous sent command packet
        double m_time_of_previous_command_packet;
        // Current state of the ardupilot
        ArduPilotState m_state;
        
        
        Task(const std::string& name, Tasks::Context& ctx):
          Tasks::Task(name, ctx),
          ref_lat(0.0),
          ref_lon(0.0),
          ref_hei(0.0),
          m_sysid(1),
          m_lat(0.0),
          m_lon(0.0),
          m_alt(0.0),
          m_external(true),
          m_current_wp(0),
          m_critical(false),
          m_cloops(0),
          m_desired_radius(0),
          m_gnd_speed(0),
          m_mode(0),
          m_changing_wp(false),
          m_error_missing(false),
          m_error_ext(false),
          m_has_setup_rate(false),
          m_vehicle_type(VEHICLE_UNKNOWN),
          m_ardu_tracker(false),
          m_ardu_vel_ctrl(true),
          m_ardu_vel_ctrl_body(true),
          m_auto_custom_maneuver(false),
          m_time_of_previous_command_packet(0)

        {
          param("Communications Timeout", m_args.comm_timeout)
          .minimumValue("1")
          .maximumValue("60")
          .defaultValue("10")
          .units(Units::Second)
          .description("Ardupilot communications timeout");

          param("Ardupilot Tracker", m_args.ardu_tracker)
          .defaultValue("false")
          .description("Use Ardupilot's waypoint tracker");

          param("Ardupilot Velocity Control", m_args.ardu_vel_ctrl)
          .defaultValue("true")
          .description("Use velocity-controller on ardupilot");

          param("Ardupilot Velocity BODY", m_args.ardu_vel_ctrl_body)
          .defaultValue("true")
          .description("Use body velocity-controller on ardupilot");

          param("Auto Custom Maneuver", m_args.auto_custom_maneuver)
          .defaultValue("false")
          .description("Automatically dispatch a custom-maneuver on DUNE-mode");

          param("Auto Custom Maneuver - Name", m_args.auto_custom_maneuver_name)
          .defaultValue("Custom Maneuver")
          .description("Name of dispatched maneuver");

          param("Auto Custom Maneuver - Plan Name", m_args.auto_custom_maneuver_plan_name)
          .defaultValue("Quick Plan")
          .description("Name of dispatched maneuver");

          param("Auto Custom Maneuver - Parameter", m_args.auto_custom_maneuver_parameter)
          .defaultValue("Some parameter")
          .description("Name of dispatched maneuver");


          param("TCP - Port", m_args.TCP_port)
          .defaultValue("5760")
          .description("Port for connection to Ardupilot");

          param("TCP - Address", m_args.TCP_addr)
          .defaultValue("127.0.0.1")
          .description("Address for connection to Ardupilot");

          param("Telemetry Rate", m_args.trate)
          .defaultValue("10")
          .units(Units::Hertz)
          .description("Telemetry output rate from Ardupilot");

          param("Default altitude", m_args.alt)
          .defaultValue("30.0")
          .units(Units::Meter)
          .description("Altitude to be used if desired Z has no units");

          param("Default speed", m_args.speed)
          .defaultValue("18.0")
          .units(Units::MeterPerSecond)
          .description("Speed to be used if desired speed is not specified");

          param("uBlox GPS", m_args.ublox)
          .defaultValue("false")
          .description("The installed GPS is uBlox");

          param("Default loiter radius", m_args.lradius)
          .defaultValue("0.0")
          .units(Units::Meter)
          .description("Loiter radius used in LoiterHere (idle)");

          param("Loitering tolerance", m_args.ltolerance)
          .defaultValue("1")
          .units(Units::Meter)
          .description("Distance to consider loitering (radius + tolerance)");

          param("Power Module", m_args.pwrm)
          .defaultValue("true")
          .description("There is a Power Module installed");

          param("Seconds before Waypoint", m_args.secs)
          .defaultValue("0.1")
          .units(Units::Second)
          .description("Seconds before actually reaching Waypoint that it is considered as reached");

          param("Meters before Waypoint", m_args.wptolerance)
                    .defaultValue("10")
                    .units(Units::Meter)
                    .description("Meters before actually reaching Waypoint that it is considered as reached");
          // Added for flight test to be able to switch between ways of switching waypoint
          param("Waypoints switch use new method",m_args.wptmethod)
          .defaultValue("true")
          .description("True if use the circle of acceptance way to switch wp, false if use the old implemented way");


          param("Minimum Command Interval", m_args.min_command_interval)
          .defaultValue("0.1")
          .units(Units::Second)
          .description("Seconds of minimum interval time to wait between each seperate desired command msg in dune mode");
          // Setup packet handlers
          // IMPORTANT: set up function to handle each type of MAVLINK packet here
          m_mlh[MAVLINK_MSG_ID_ATTITUDE] = &Task::handleAttitudePacket;
          m_mlh[MAVLINK_MSG_ID_GLOBAL_POSITION_INT] = &Task::handlePositionPacket;
          m_mlh[MAVLINK_MSG_ID_HWSTATUS] = &Task::handleHWStatusPacket;
          m_mlh[MAVLINK_MSG_ID_SCALED_PRESSURE] = &Task::handleScaledPressurePacket;
          m_mlh[MAVLINK_MSG_ID_GPS_RAW_INT] = &Task::handleRawGPSPacket;
          m_mlh[MAVLINK_MSG_ID_WIND] = &Task::handleWindPacket;
          m_mlh[MAVLINK_MSG_ID_COMMAND_ACK] = &Task::handleCmdAckPacket;
          m_mlh[MAVLINK_MSG_ID_MISSION_ACK] = &Task::handleMissionAckPacket;
          m_mlh[MAVLINK_MSG_ID_MISSION_CURRENT] = &Task::handleMissionCurrentPacket;
          m_mlh[MAVLINK_MSG_ID_STATUSTEXT] = &Task::handleStatusTextPacket;
          m_mlh[MAVLINK_MSG_ID_HEARTBEAT] = &Task::handleHeartbeatPacket;
          m_mlh[MAVLINK_MSG_ID_NAV_CONTROLLER_OUTPUT] = &Task::handleNavControllerPacket;
          m_mlh[MAVLINK_MSG_ID_MISSION_ITEM] = &Task::handleMissionItemPacket;
          m_mlh[MAVLINK_MSG_ID_SYS_STATUS] = &Task::handleSystemStatusPacket;
          m_mlh[MAVLINK_MSG_ID_VFR_HUD] = &Task::handleHUDPacket;
          m_mlh[MAVLINK_MSG_ID_SYSTEM_TIME] = &Task::handleSystemTimePacket;


          // Setup processing of IMC messages
          bind<DesiredPath>(this);
          //bind<DesiredExtCtrl>(this);
          bind<SetServoPosition>(this);
          bind<IdleManeuver>(this);
          bind<ControlLoops>(this);
          bind<PowerChannelControl>(this);
          bind<IMC::DesiredVelocity>(this);
          bind<IMC::DesiredYaw>(this);
          bind<IMC::DesiredZ>(this);
          bind<IMC::DesiredRoll>(this);
          bind<IMC::DesiredPitch>(this);
          bind<IMC::DesiredControl>(this);
          //bind<IMC::EstimatedState>(this);
//          bind<VehicleMedium>(this);

          // Misc. initialization
          m_last_pkt_time = 0; // time of last packet from Ardupilot
          m_estate.clear();
          for( int i = 0; i < ACL_MAX; i++)
          {
            m_updated_desired[i] = false;
          }

          m_state.armed = false;
          m_state.mode = "";
          m_state.type = IMC::ArduPilotState::AS_STATE;
        }
        //! Update internal parameters.
        void
        onUpdateParameters(void)
        {
          // Update allowable parameters
          if (paramChanged(m_args.auto_custom_maneuver)) {
            m_auto_custom_maneuver = m_args.auto_custom_maneuver;
          }
          if (paramChanged(m_args.ardu_vel_ctrl_body)) {
            m_ardu_vel_ctrl_body = m_args.ardu_vel_ctrl_body;
          }
          // Test
         // TupleList
          //inf("Got list: %s", m_args.auto_custom_maneuver_parameter.)
        }

        void
        onResourceRelease(void)
        {
          Memory::clear(m_TCP_sock);
        }

        void
        onResourceAcquisition(void)
        {
          m_TCP_addr = m_args.TCP_addr;
          m_TCP_port = m_args.TCP_port;
          openConnection();
        }

        //! Initialize resources
        void
        onResourceInitialization(void)
        {

          // Only allow this from ini-file.
          m_ardu_tracker  = m_args.ardu_tracker;
          m_ardu_vel_ctrl = m_args.ardu_vel_ctrl;
        }

        void
        openConnection(void)
        {
          try
          {
            m_TCP_sock = new TCPSocket;
            m_TCP_sock->connect(m_TCP_addr, m_TCP_port);
            setupRate(m_args.trate);
            inf(DTR("Ardupilot interface initialized"));
          }
          catch (...)
          {
            m_TCP_sock = 0;
            war(DTR("Connection failed, retrying..."));
            setEntityState(IMC::EntityState::ESTA_NORMAL, Status::CODE_COM_ERROR);
          }
        }

        void
        setupRate(uint8_t rate)
        {
          uint8_t buf[512];
          mavlink_message_t* msg = new mavlink_message_t;

          mavlink_msg_request_data_stream_pack(255, 0, msg,
                                               m_sysid,
                                               0,
                                               MAV_DATA_STREAM_EXTRA1,
                                               rate,
                                               1);

          uint16_t n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          spew("ATTITUDE Stream setup to %d Hertz", rate);

          mavlink_msg_request_data_stream_pack(255, 0, msg,
                                               m_sysid,
                                               0,
                                               MAV_DATA_STREAM_EXTRA2,
                                               rate,
                                               1);

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          spew("VFR Stream setup to %d Hertz", rate);

          mavlink_msg_request_data_stream_pack(255, 0, msg,
                                               m_sysid,
                                               0,
                                               MAV_DATA_STREAM_POSITION,
                                               rate,
                                               1);

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          spew("POSITION Stream setup to %d Hertz", rate);

          mavlink_msg_request_data_stream_pack(255, 0, msg,
                                               m_sysid,
                                               0,
                                               MAV_DATA_STREAM_EXTENDED_STATUS,
                                               (int)(rate/5),
                                               1);

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          spew("STATUS Stream setup to %d Hertz", (int)(rate/5));

          mavlink_msg_request_data_stream_pack(255, 0, msg,
                                               m_sysid,
                                               0,
                                               MAV_DATA_STREAM_EXTRA3,
                                               1,
                                               1);

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          spew("AHRS-HWSTATUS-WIND Stream setup to 1 Hertz");

          mavlink_msg_request_data_stream_pack(255, 0, msg,
                                               m_sysid,
                                               0,
                                               MAV_DATA_STREAM_RAW_SENSORS,
                                               1,
                                               1);

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          spew("SENSORS Stream setup to 1 Hertz");

          mavlink_msg_request_data_stream_pack(255, 0, msg,
                                               m_sysid,
                                               0,
                                               MAV_DATA_STREAM_RC_CHANNELS,
                                               1,
                                               0);

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          spew("RC Stream disabled");
        }

        void
        info(uint32_t was, uint32_t is, uint32_t loop, const char* desc)
        {
          was &= loop;
          is &= loop;

          if (was && !is)
            debug("%s - deactivating", desc);
          else if (!was && is)
            debug("%s - activating", desc);
        }

        /*
         * Main methods for sending new stuff to ardupilot.
         */
        void
        consume(const IMC::DesiredYaw* yaw)
        {
          if (!(m_cloops & IMC::CL_YAW))
          {
           debug("yaw control of APM is NOT active");
           return;
          }

          // Update internal variable for yaw control
          m_d_yaw = *yaw;

          update(ACL_YAW);

        }



        void
        consume(const IMC::DesiredZ* alt)
        {
          if (!(m_cloops & IMC::CL_ALTITUDE))
          {
           debug("altitude control of APM is NOT active");
           return;
          }

          // Update internal variable for alt control
          if(alt->z_units == Z_ALTITUDE)
          {
            //m_extCtrl.z = alt->value;
            //m_extCtrl.z_units = alt->z_units;
            m_d_z = *alt;
            update(ACL_ALTITUDE);
          }
        }

        /*
         * Main workhorse
         */


        void
        consume(const IMC::DesiredVelocity* vel)
        {
          // Check if is enabled.
          if (!(m_cloops & IMC::CL_VELOCITY))
          {
           debug("velocity control of APM is NOT active");
           return;
          }

          m_d_velocity = *vel;
          update(ACL_VELOCITY);
        }

        void
        consume(const IMC::DesiredRoll* roll)
        {
          // Check if is enabled.
          if (!(m_cloops & IMC::CL_ROLL))
          {
           debug("Roll control of APM is NOT active");
           return;
          }

          m_d_roll = *roll;
          update(ACL_ROLL);
        }

        void
        consume(const IMC::DesiredPitch* pitch)
        {
          // Check if is enabled.
          if (!(m_cloops & IMC::CL_PITCH))
          {
           debug("Pitch control of APM is NOT active");
           return;
          }

          m_d_pitch = *pitch;
          update(ACL_PITCH);
        }

        void
        consume(const IMC::DesiredControl* control)
        {
          // Check if is enabled.
          if (!(m_cloops & IMC::CL_FORCE))
          {
           debug("Force control of APM is NOT active");
           return;
          }

          m_d_control = *control;
          update(ACL_CONTROL);
        }

        /*
         * Marks a new reference as updated, and checks if it should be sent.
         */
        bool
        update(APM_c_loops cloop)
        {
          bool send = false;
          if (cloop >= ACL_MAX)
            return false;

          m_updated_desired[cloop] = true;

          // If not min interval time since last send, no send.
          double cur_time = Clock::get();

          if( m_time_of_previous_command_packet + m_args.min_command_interval > cur_time)
            return false;

          // Roll and Pitch are connected. Either both or neither must be updated.
          // aka, only send both at the same time.

          uint8_t target = 0;
          float trans_x = 0, trans_y = 0, trans_z = 0,
              roll    = 0, pitch   = 0, yaw     = 0;


          if( m_updated_desired[ACL_YAW])
          {
            target |= TARGET_YAW_SET;
            yaw    = m_d_yaw.value;

            // Mark as sent and ready
            send = true;
            m_updated_desired[ACL_YAW] = false;
          }
          if( m_updated_desired[ACL_ROLL] && m_updated_desired[ACL_PITCH])
          {
            target |= TARGET_RP_SET;
            roll   = m_d_roll.value;
            pitch  = m_d_pitch.value;

            // Mark as sent and ready
            send = true;
            m_updated_desired[ACL_ROLL] = false;
            m_updated_desired[ACL_PITCH]= false;

          }
          if( m_updated_desired[ACL_CONTROL])
          {
            // Just Z thrust for now.
            if ( m_d_control.flags & IMC::DesiredControl::FL_Z)
            {
              target |= TARGET_Z_THRUST;
              trans_z = m_d_control.z;

              // Mark as sent and ready
              send = true;
              m_updated_desired[ACL_CONTROL] = false;
            }
          }
          if( m_updated_desired[ACL_ALTITUDE])
          {
            target &= ~ TARGET_MASK_Z;
            target |=   TARGET_Z_ALT;
            trans_z = m_d_z.value;

            // Mark as sent and ready
            send = true;
            m_updated_desired[ACL_ALTITUDE] = false;
          }
          if( m_updated_desired[ACL_VERTICAL_RATE])
          {

            if ( m_d_velocity.flags & IMC::DesiredVelocity::FL_HEAVE)
            {

              // TODO: CHECK DIRECTION ON APM!
              target &= ~ TARGET_MASK_Z;
              target |=   TARGET_Z_VEL;
              trans_z = m_d_velocity.w;

              // Mark as sent and ready
              send = true;
              m_updated_desired[ACL_VERTICAL_RATE] = false;
            }
          }
          if( m_updated_desired[ACL_VELOCITY])
          {
            if ( (m_d_velocity.flags & IMC::DesiredVelocity::FL_SURGE) &&
                (m_d_velocity.flags & IMC::DesiredVelocity::FL_SWAY) )
            {
              // Disable roll-pitch
              target &= ~ TARGET_MASK_RP;
              target |=   TARGET_RP_NONE;

              // Set vel
              if (m_ardu_vel_ctrl_body) {
                target |= TARGET_XY_VEL_BODY;
              } else {
                target |= TARGET_XY_VEL_NED;
              }
              trans_x = m_d_velocity.u;
              trans_y = m_d_velocity.v;

              // Mark as sent and ready
              send = true;
              m_updated_desired[ACL_VELOCITY] = false;
            }
          }




          if (send)
          {
            // Update previous time
            m_time_of_previous_command_packet = cur_time;

            mavlink_message_t* msg = new mavlink_message_t;
            mavlink_msg_setpoint_6dof_pack(255, 0, msg,
                target,
                trans_x, // Trans x
                trans_y, // Trans y
                trans_z, // Trans z
                roll, // roll
                pitch, // pitch
                yaw); //yaw
            uint16_t n = mavlink_msg_to_send_buffer(m_buf, msg);
            sendData(m_buf, n);
            debug("Desired Velocity packet sent to Ardupilot.");
            debug("X: %0.2f Y: %0.2f Z: %0.2f Yaw: %0.2f", trans_x, trans_y, trans_z, yaw);
          }

          return send;
        }


        /*
         * May set dune-mode on arudpilot
         * BUT, do not do anything IF external or critical.
         */
        void
        consume(const IMC::ControlLoops* cloops)
        {
          if(m_external || m_critical)
            return;

          uint32_t prev = m_cloops;

          if (cloops->enable)
          {
            m_cloops |= cloops->mask;
            // Path tracker can only be enabled if ardu-tracker is on.
            if((!m_args.ardu_tracker) && (cloops->mask & IMC::CL_PATH))
            {
              debug("tracker is NOT enabled");
              m_cloops &= ~IMC::CL_PATH;
            }
            
            //if(!(m_args.ardu_tracker) && (cloops->mask & IMC::CL_EXTERNAL) )
            if( (cloops->mask & IMC::CL_GUIDANCE) )
            {
              uint8_t buf[512];
              mavlink_message_t* msg = new mavlink_message_t;

              mavlink_msg_set_mode_pack(255, 0, msg,
                                        m_sysid,
                                        1,
                                        12); //! DUNE mode on arducopter is 12

              uint16_t n = mavlink_msg_to_send_buffer(buf, msg);
              sendData(buf, n);
              debug("DUNE MODE on ardupilot is set");
            }
          }
          else
            m_cloops &= ~cloops->mask;

          //info(prev, m_cloops, IMC::CL_SPEED, "speed control");
          info(prev, m_cloops, IMC::CL_EXTERNAL, "external control");
          info(prev, m_cloops, IMC::CL_GUIDANCE, "external control");
          info(prev, m_cloops, IMC::CL_PATH, "path control");
          info(prev, m_cloops, IMC::CL_VELOCITY, "velocity control");
          info(prev, m_cloops, IMC::CL_VERTICAL_RATE, "Vertical Rate control");
          info(prev, m_cloops, IMC::CL_ROLLPITCH, "roll-pitch control");
        }
        /*
        void
        consume(const IMC::DesiredExtCtrl* d_extCtrl)
        {
          if (!(m_cloops & IMC::CL_EXTERNAL))
          {
            debug("external control of APM is NOT active");
            return;
          }
          
          m_extCtrl.roll  = d_extCtrl->roll;
          m_extCtrl.pitch = d_extCtrl->pitch;
          m_extCtrl.yaw   = d_extCtrl->yaw;
          m_extCtrl.z     = d_extCtrl->z;

          uint8_t buf[512];

          mavlink_message_t* msg = new mavlink_message_t;
          mavlink_msg_set_roll_pitch_yaw_thrust_pack(255, 0, msg,
                                                     1,
                                                     1,
                                                     (float)m_extCtrl.roll,
                                                     (float)m_extCtrl.pitch,
                                                     (float)m_extCtrl.yaw,
                                                     (float)m_extCtrl.z);
          uint16_t n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);
          debug("DesiredExtCtrl packet sent to Ardupilot.");
          debug("R: %0.2f P: %0.2f Y: %0.2f Z: %0.2f", m_extCtrl.roll, m_extCtrl.pitch, m_extCtrl.yaw, m_extCtrl.z);
        }

*/
        /*
         * This basically transfers internal dune paths to ardupilot wp.
         * Is disabled if ardu_tracker is disabled.
         */
        void
        consume(const IMC::DesiredPath* path)
        {
          if(!m_args.ardu_tracker)
            return;

          if(!(m_cloops & IMC::CL_PATH))
          {
            debug("path control is NOT active");
            return;
          }

          uint8_t buf[512];

          mavlink_message_t* msg = new mavlink_message_t;

          mavlink_msg_param_set_pack(255, 0, msg,
                                     m_sysid, //! target_system System ID
                                     0, //! target_component Component ID
                                     "TRIM_ARSPD_CM", //! Parameter name
                                     (int)(path->speed * 100), //! Parameter value
                                     MAV_PARAM_TYPE_INT16); //! Parameter type

          int n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);

          mavlink_msg_param_set_pack(255, 0, msg,
                                     m_sysid, //! target_system System ID
                                     0, //! target_component Component ID
                                     "WP_LOITER_RAD", //! Parameter name
                                     path->flags & DesiredPath::FL_CCLOCKW ? (-1 * path->lradius) : (path->lradius), //! Parameter value
                                     MAV_PARAM_TYPE_INT16); //! Parameter type

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);

          m_desired_radius = (uint16_t) path->lradius;

          mavlink_msg_mission_count_pack(255, 0, msg,
                                         m_sysid, //! target_system System ID
                                         0, //! target_component Component ID
                                         3); //! size of Mission

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);

          mavlink_msg_mission_write_partial_list_pack(255, 0, msg,
                                                      m_sysid, //! target_system System ID
                                                      0, //! target_component Component ID
                                                      1, //! start_index Start index, 0 by default and smaller / equal to the largest index of the current onboard list
                                                      1); //! end_index End index, equal or greater than start index

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);

          float alt = (path->end_z_units & IMC::Z_NONE) ? m_args.alt : (float)path->end_z;

          //! Destination
          mavlink_msg_mission_item_pack(255, 0, msg,
                                        m_sysid, //! target_system System ID
                                        0, //! target_component Component ID
                                        1, //! seq Sequence
                                        MAV_FRAME_GLOBAL, //! frame The coordinate system of the MISSION. see MAV_FRAME in mavlink_types.h
                                        MAV_CMD_NAV_LOITER_UNLIM, //! command The scheduled action for the MISSION. see MAV_CMD in ardupilotmega.h
                                        2, //! current false:0, true:1
                                        0, //! autocontinue to next wp
                                        0, //! Not used
                                        0, //! Not used
                                        path->flags & DesiredPath::FL_CCLOCKW ? -1 : 0, //! If <0, then CCW loiter
                                        0, //! Not used
                                        (float)Angles::degrees(path->end_lat), //! x PARAM5 / local: x position, global: latitude
                                        (float)Angles::degrees(path->end_lon), //! y PARAM6 / y position: global: longitude
                                        alt);//! z PARAM7 / z position: global: altitude

          n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);

          m_changing_wp = true;

          m_pcs.start_lat = Angles::radians(m_lat);
          m_pcs.start_lon = Angles::radians(m_lon);
          m_pcs.start_z = m_alt;
          m_pcs.start_z_units = IMC::Z_HEIGHT;

          m_pcs.end_lat = path->end_lat;
          m_pcs.end_lon = path->end_lon;
          m_pcs.end_z = alt;
          m_pcs.end_z_units = IMC::Z_HEIGHT;
          m_pcs.flags = PathControlState::FL_3DTRACK | PathControlState::FL_CCLOCKW;
          m_pcs.flags &= path->flags;
          m_pcs.lradius = path->lradius;

          dispatch(m_pcs);

          debug("Waypoint packet sent to Ardupilot");
        }

        /*
         * Used as a "wait" command, if no new maneouver in the plan is avaiable.
         */

        void
        loiterHere(void)
        {

          if(getEntityState() != IMC::EntityState::ESTA_NORMAL)
            return;

          mavlink_message_t* msg = new mavlink_message_t;
          uint8_t buf[512];

          mavlink_msg_param_set_pack(255, 0, msg,
                                     m_sysid, //! target_system System ID
                                     0, //! target_component Component ID
                                     "WP_LOITER_RAD", //! Parameter name
                                     m_args.lradius, //! Parameter value
                                     MAV_PARAM_TYPE_INT16); //! Parameter type

          int n = mavlink_msg_to_send_buffer(buf, msg);
          sendData(buf, n);

          sendCommandPacket(MAV_CMD_NAV_LOITER_UNLIM);
          debug("Sent LOITER packet to Ardupilot");

          m_pcs.start_lat = m_fix.lat;
          m_pcs.start_lon = m_fix.lon;
          m_pcs.start_z = m_fix.height;

          m_pcs.end_lat = m_fix.lat;
          m_pcs.end_lon = m_fix.lon;
          m_pcs.end_z = m_fix.height;

          m_pcs.flags = PathControlState::FL_LOITERING | PathControlState::FL_NEAR | (m_args.lradius < 0 ? PathControlState::FL_CCLOCKW : 0);
          m_pcs.lradius = m_args.lradius * (m_args.lradius < 0 ? -1 : 1);

          dispatch(m_pcs);
        }


        /*
         * Trigger for loiterHere.
         */
        void
        consume(const IMC::IdleManeuver* idle)
        {
          (void)idle;

          loiterHere();
        }

        /*
         * Low-level set servo position for external servos.
         */

        void
        consume(const IMC::SetServoPosition* servo)
        {
          sendCommandPacket(MAV_CMD_DO_SET_SERVO,
                            servo->id, // Servo number
                            servo->value, // PWM (microseconds, 1000 to 2000 typical)
                            0, // Empty
                            0); //Empty
          spew("SetServo packet sent to Ardupilot");
        }

        /*
         * Dunno what this is.
         */

        void
        consume(const IMC::PowerChannelControl* pcc)
        {
          trace("Trigger Request Received");

          if (pcc->op & IMC::PowerChannelControl::PCC_OP_TURN_ON)
            sendCommandPacket(MAV_CMD_DO_SET_RELAY, 1);
          else
            sendCommandPacket(MAV_CMD_DO_SET_RELAY, 0);
        }

//        void
//        consume(const IMC::VehicleMedium* vm)
//        {
//          m_ground = (vm->medium == IMC::VehicleMedium::VM_GROUND);
//        }

        void
        sendCommandPacket(uint16_t cmd, float arg1=0, float arg2=0, float arg3=0, float arg4=0, float arg5=0, float arg6=0, float arg7=0)
        {
          uint8_t buf[512];

          mavlink_message_t msg;

          trace("%0.2f %0.2f %0.2f %0.2f %0.2f %0.2f %0.2f", arg1, arg2, arg3, arg4, arg5, arg6, arg7);

          mavlink_msg_command_long_pack(255, 0, &msg, m_sysid, 0, cmd, 0, arg1, arg2, arg3, arg4, arg5, arg6, arg7);

          uint16_t n = mavlink_msg_to_send_buffer(buf, &msg);
          sendData(buf, n);
        }

        void
        onMain(void)
        {

          bool prev_external = m_external;

          while (!stopping())
          {
            // Update previous
            prev_external = m_external;

            // Handle data
            if(m_TCP_sock)
            {
              handleArdupilotData();
            }
            else
            {
              Time::Delay::wait(0.5);
              openConnection();
            }

            // Check if auto trigger manueuver
            if(m_auto_custom_maneuver)
            {
              if(m_external && !prev_external)
              {

                // Suddenly in external
                IMC::GuidanceControlState stop;
                stop.flags = IMC::GuidanceControlState::FL_DONE;
                dispatch(stop);
              }
              if(prev_external && !m_external)
              {
                // Was external, but not now!

                // Create maneuver
                IMC::CustomManeuver man;
                man.name ="My awesome maneuver!";

                // Create plan
                IMC::PlanControl plan;
                plan.op = IMC::PlanControl::PC_START;
                plan.type = IMC::PlanControl::PC_REQUEST;
                plan.plan_id = "Test plan";
                plan.request_id = 10;

                // Apply custom maneuver to plan
                IMC::InlineMessage<IMC::Message> arg;
                arg.set(man);
                plan.arg = arg;
                err("Custom plan activated");
                dispatch(plan);
              }
            }
            if(m_external || m_critical)
            {
              if(getEntityState() != IMC::EntityState::ESTA_ERROR && !m_error_ext)
              {
                setEntityState(IMC::EntityState::ESTA_ERROR, "External Control");
                m_error_ext = true;
              }
            }
            else if(getEntityState() != IMC::EntityState::ESTA_NORMAL)
            {
              setEntityState(IMC::EntityState::ESTA_NORMAL, Status::CODE_ACTIVE);
              m_error_ext = false;
            }

            // Handle IMC messages from bus
            consumeMessages();

            // update previous external
          }
        }

        bool
        poll(double timeout)
        {
          if (m_TCP_sock != NULL)
            return Poll::poll(*m_TCP_sock, timeout);

          return false;
        }

        int
        sendData(uint8_t* bfr, int size)
        {
          if (m_TCP_sock)
          {
            trace("Sending something");
            return m_TCP_sock->write((char*)bfr, size);
          }
          return 0;
        }

        int
        receiveData(uint8_t* buf, size_t blen)
        {
          if (m_TCP_sock)
          {
            try
            {
              return m_TCP_sock->read(buf, blen);
            }
            catch (std::runtime_error& e)
            {
              err("%s", e.what());
              war(DTR("Connection lost, retrying..."));
              Memory::clear(m_TCP_sock);

              m_TCP_sock = new Network::TCPSocket;
              m_TCP_sock->connect(m_TCP_addr, m_TCP_port);
              return 0;
            }
          }
          return 0;
        }

        void
        handleArdupilotData(void)
        {
          mavlink_status_t status;

          double now = Clock::get();

          while (poll(0.01))
          {
            int n = receiveData(m_buf, sizeof(m_buf));

            if (n < 0)
            {
              debug("Receive error");
              break;
            }


            for (int i = 0; i < n; i++)
            {
              int rv = mavlink_parse_char(MAVLINK_COMM_0, m_buf[i], &m_msg, &status);
              if (status.packet_rx_drop_count)
              {
                switch(status.parse_state)
                {
                  case MAVLINK_PARSE_STATE_IDLE:
                    spew("failed at state IDLE");
                    break;
                  case MAVLINK_PARSE_STATE_GOT_STX:
                    spew("failed at state GOT_STX");
                    break;
                  case MAVLINK_PARSE_STATE_GOT_LENGTH:
                    spew("failed at state GOT_LENGTH");
                    break;
                  case MAVLINK_PARSE_STATE_GOT_SYSID:
                    spew("failed at state GOT_SYSID");
                    break;
                  case MAVLINK_PARSE_STATE_GOT_COMPID:
                    spew("failed at state GOT_COMPID");
                    break;
                  case MAVLINK_PARSE_STATE_GOT_MSGID:
                    spew("failed at state GOT_MSGID");
                    break;
                  case MAVLINK_PARSE_STATE_GOT_PAYLOAD:
                    spew("failed at state GOT_PAYLOAD");
                    break;
                  case MAVLINK_PARSE_STATE_GOT_CRC1:
                    spew("failed at state GOT_CRC1");
                    break;
                  default:
                    spew("failed OTHER");
                }
              }
              if (rv)
              {
                switch ((int)m_msg.msgid)
                {
                  default:
                    trace("UNDEF: %u", m_msg.msgid);
                    break;
                  case MAVLINK_MSG_ID_HEARTBEAT:
                    trace("HEARTBEAT");
                    break;
                  case MAVLINK_MSG_ID_SYS_STATUS:
                    trace("SYS_STATUS");
                    break;
                  case 22:
                    trace("PARAM_VALUE");
                    break;
                  case MAVLINK_MSG_ID_GPS_RAW_INT:
                    spew("GPS_RAW");
                    break;
                  case 27:
                    trace("IMU_RAW");
                    break;
                  case MAVLINK_MSG_ID_SCALED_PRESSURE:
                    spew("SCALED_PRESSURE");
                    break;
                  case MAVLINK_MSG_ID_ATTITUDE:
                    spew("ATTITUDE");
                    break;
                  case MAVLINK_MSG_ID_GLOBAL_POSITION_INT:
                    spew("GLOBAL_POSITION_INT");
                    break;
                  case 34:
                    trace("RC_CHANNELS_SCALED");
                    break;
                  case 35:
                    trace("RC_CHANNELS_RAW");
                    break;
                  case MAVLINK_MSG_ID_MISSION_ITEM:
                    trace("MISSION_ITEM");
                    break;
                  case MAVLINK_MSG_ID_MISSION_CURRENT:
                    trace("MISSION_CURRENT");
                    break;
                  case 44:
                    trace("MISSION_COUNT");
                    break;
                  case MAVLINK_MSG_ID_MISSION_ACK:
                    spew("MISSION_ACK");
                    break;
                  case MAVLINK_MSG_ID_NAV_CONTROLLER_OUTPUT:
                    trace("NAV_CONTROLLER_OUTPUT");
                    break;
                  case MAVLINK_MSG_ID_VFR_HUD:
                    trace("VFR_HUD");
                    break;
                  case MAVLINK_MSG_ID_COMMAND_ACK:
                    spew("CMD_ACK");
                    break;
                  case MAVLINK_MSG_ID_BATTERY_STATUS:
                    spew("BATTERY_STAT");
                    break;
                  case 150:
                    trace("SENSOR_OFFSETS");
                    break;
                  case 152:
                    trace("MEMINFO");
                    break;
                  case 162:
                    trace("FENCE_STATUS");
                    break;
                  case 163:
                    trace("AHRS");
                    break;
                  case 164:
                    trace("SIM_STATE");
                    break;
                  case MAVLINK_MSG_ID_HWSTATUS:
                    spew("HW_STATUS");
                    break;
                  case MAVLINK_MSG_ID_WIND:
                    spew("WIND");
                    break;
                  case MAVLINK_MSG_ID_STATUSTEXT:
                    trace("STATUSTEXT");
                    break;
                }

                PktHandler h = m_mlh[m_msg.msgid];

                if (!h)
                  continue;  // Ignore this packet (no handler for it)

                // Call handler
                (this->*h)(&m_msg);
                m_sysid = m_msg.sysid;

                m_last_pkt_time = now;
              }
            }
          }

          m_critical = false;

          if ((now - m_last_pkt_time >= m_args.comm_timeout) && !m_error_missing)
          {
            setEntityState(IMC::EntityState::ESTA_ERROR, Status::CODE_MISSING_DATA);
            m_critical = true;
            m_error_missing = true;
          }
          else
            m_error_missing = false;
        }

        void
        handleAttitudePacket(const mavlink_message_t* msg)
        {
          mavlink_attitude_t att;

          mavlink_msg_attitude_decode(msg, &att);
          m_estate.phi = att.roll;
          m_estate.theta = att.pitch;
          m_estate.psi = att.yaw;
          m_estate.p = att.rollspeed;
          m_estate.q = att.pitchspeed;
          m_estate.r = att.yawspeed;
        }

        void
        handlePositionPacket(const mavlink_message_t* msg)
        {
          mavlink_global_position_int_t gp;
          mavlink_msg_global_position_int_decode(msg, &gp);

          double lat = Angles::radians((double)gp.lat * 1e-07);
          double lon = Angles::radians((double)gp.lon * 1e-07);
          float hei = (float)gp.alt * 1e-03;

          m_lat = (double)gp.lat * 1e-07;
          m_lon = (double)gp.lon * 1e-07;
          m_alt = (float)gp.alt * 1e-03;

          double distance_to_ref = WGS84::distance(ref_lat,ref_lon,ref_hei,
                                                   lat,lon,hei);

          if (distance_to_ref>1000)
          {
            m_estate.lat = lat;
            m_estate.lon = lon;
            m_estate.height = hei;

            m_estate.x = 0;
            m_estate.y = 0;
            m_estate.z = 0;

            ref_lat = lat;
            ref_lon = lon;
            ref_hei = hei;
          }
          else
          {
            WGS84::displacement(ref_lat, ref_lon, ref_hei,
                                lat, lon, hei,
                                &m_estate.x, &m_estate.y, &m_estate.z);

            m_estate.lat = ref_lat;
            m_estate.lon = ref_lon;
            m_estate.height = ref_hei;
          }

          m_estate.vx = 1e-02 * gp.vx;
          m_estate.vy = 1e-02 * gp.vy;
          m_estate.vz = -1e-02 * gp.vz;

          // Note: the following will yield body-fixed *ground* velocity
          // Maybe this can be fixed w/IAS readings (anyway not too important)
          BodyFixedFrame::toBodyFrame(m_estate.phi, m_estate.theta, m_estate.psi,
                                      m_estate.vx, m_estate.vy, m_estate.vz,
                                      &m_estate.u, &m_estate.v, &m_estate.w);

          m_estate.depth = -1;
          m_estate.alt = gp.relative_alt * 1e-03;

          dispatch(m_estate);
        }

        void
        handleHWStatusPacket(const mavlink_message_t* msg)
        {
          if(m_args.pwrm)
          {
            (void) msg;
            return;
          }

          mavlink_hwstatus_t hw_status;

          mavlink_msg_hwstatus_decode(msg, &hw_status);

          m_volt.value = 0.001 * hw_status.Vcc;

          dispatch(m_volt);
        }

        void
        handleSystemStatusPacket(const mavlink_message_t* msg)
        {
          if (!m_args.pwrm)
          {
            (void) msg;
            return;
          }
          mavlink_sys_status_t sys_status;

          mavlink_msg_sys_status_decode(msg, &sys_status);

          m_volt.value = 0.001 * (float)sys_status.voltage_battery;
          m_curr.value = 0.01 * (float)sys_status.current_battery;
          m_fuel.value = (float)sys_status.battery_remaining;

          dispatch(m_volt);
          dispatch(m_curr);
          dispatch(m_fuel);
        }

        void
        handleScaledPressurePacket(const mavlink_message_t* msg)
        {
          mavlink_scaled_pressure_t sc_press;

          mavlink_msg_scaled_pressure_decode(msg, &sc_press);

          m_pressure.value = sc_press.press_abs;
          m_temp.value = 0.01 * sc_press.temperature;

          dispatch(m_pressure);
          dispatch(m_temp);
        }

        void
        handleRawGPSPacket(const mavlink_message_t* msg)
        {
          mavlink_gps_raw_int_t gps_raw;

          mavlink_msg_gps_raw_int_decode(msg, &gps_raw);

          m_fix.cog = Angles::radians((double)gps_raw.cog * 0.01);
          m_fix.sog = (float)gps_raw.vel * 0.01;
          m_fix.hdop = (float)gps_raw.eph * 0.01;
          m_fix.vdop = (float)gps_raw.epv * 0.01;
          m_fix.lat = Angles::radians((double)gps_raw.lat * 1e-07);
          m_fix.lon = Angles::radians((double)gps_raw.lon * 1e-07);
          m_fix.height = (double)gps_raw.alt * 0.001;
          m_fix.satellites = gps_raw.satellites_visible;

          m_fix.validity = 0;
          if (gps_raw.fix_type>1)
            m_fix.validity |= IMC::GpsFix::GFV_VALID_POS;
          if (m_fix.utc_year>2012)
            m_fix.validity |= (IMC::GpsFix::GFV_VALID_TIME | IMC::GpsFix::GFV_VALID_DATE);
        }

        void
        handleWindPacket(const mavlink_message_t* msg)
        {
          mavlink_wind_t wind;

          mavlink_msg_wind_decode(msg, &wind);

          double wind_dir_rad = wind.direction * Math::c_pi / 180;

          m_stream.x = std::cos(wind_dir_rad) * wind.speed;
          m_stream.y = std::sin(wind_dir_rad) * wind.speed;
          m_stream.z = wind.speed_z;

          dispatch(m_stream);
        }

        void
        handleCmdAckPacket(const mavlink_message_t* msg)
        {
          mavlink_command_ack_t cmd_ack;

          mavlink_msg_command_ack_decode(msg, &cmd_ack);
          debug("Command %d was received, result is %d", cmd_ack.command, cmd_ack.result);
          m_changing_wp = false;
        }

        void
        handleMissionAckPacket(const mavlink_message_t* msg)
        {
          mavlink_mission_ack_t miss_ack;

          mavlink_msg_mission_ack_decode(msg, &miss_ack);
          debug("Mission was received, result is %d", miss_ack.type);
          m_changing_wp = false;
        }

        void
        handleMissionCurrentPacket(const mavlink_message_t* msg)
        {
          mavlink_mission_current_t miss_curr;

          mavlink_msg_mission_current_decode(msg, &miss_curr);
          m_current_wp = miss_curr.seq;
          trace("Current mission item: %d", miss_curr.seq);

          uint8_t buf[512];

          mavlink_message_t* msg_out = new mavlink_message_t;

          mavlink_msg_mission_request_pack(255, 0, msg_out,
                                           m_sysid, //! target_system System ID
                                           0, //! target_component Component ID
                                           m_current_wp); //! Mission item to request

          uint16_t n = mavlink_msg_to_send_buffer(buf, msg_out);
          sendData(buf, n);
        }

        void
        handleStatusTextPacket(const mavlink_message_t* msg)
        {
          mavlink_statustext_t stat_tex;
          mavlink_msg_statustext_decode(msg, &stat_tex);
          debug("Status: %s", stat_tex.text);
        }

        void
        handleHeartbeatPacket(const mavlink_message_t* msg)
        {
          if(!m_has_setup_rate)
          {
            m_has_setup_rate = true;
            setupRate(m_args.trate);
            debug("Rates setup second time.");
          }

          mavlink_heartbeat_t hbt;
          mavlink_msg_heartbeat_decode(msg, &hbt);

          // Update vehicle type if applicable
          if (m_vehicle_type == VEHICLE_UNKNOWN)
          {
            MAV_TYPE mav_type = static_cast<MAV_TYPE>(hbt.type);
            switch (mav_type)
            {
            default:
              err(DTR("Controlling an unknown vehicle type."));
              break;
            case MAV_TYPE_FIXED_WING:
              m_vehicle_type = VEHICLE_FIXEDWING;
              debug("Controlling a fixed-wing vehicle.");
              break;
            case MAV_TYPE_QUADROTOR:
            case MAV_TYPE_HEXAROTOR:
            case MAV_TYPE_OCTOROTOR:
            case MAV_TYPE_TRICOPTER:
              m_vehicle_type = VEHICLE_COPTER;
              debug("Controlling a multicopter.");
              break;
            }
          }


          m_state.armed = (hbt.base_mode & MAV_MODE_FLAG_SAFETY_ARMED) ? true : false;
          if(m_state.armed)
            trace("Ardupilot is ARMED");
          else
            trace("Ardupilot is NOT armed");


          m_mode = hbt.custom_mode;
          if (m_vehicle_type == VEHICLE_COPTER)
          {
            switch(m_mode)
            {
            default:
              m_external = true;
              break;
            case 3:
              trace("AUTO");
              m_external = false;
              break;
            case 5:
              trace("LOITER");
              m_external = false;
              break;
            case 12:
              trace("DUNE");
              m_external = false;
              break;
            case 4:
              trace("GUIDED");
              m_external = false;
              break;
            }

            // Again to update mode state
            switch(m_mode)
            {

              case CP_MODE_STABILIZE:
                m_state.mode="STABILIZE";
                break;
              case CP_MODE_ACRO:
                m_state.mode="ACRO";
                break;
              case CP_MODE_ALT_HOLD:
                m_state.mode="ALT_HOLD";
                break;
              case CP_MODE_AUTO:
                m_state.mode="AUTO";
                break;
              case CP_MODE_GUIDED:
                m_state.mode="GUIDED";
                break;
              case CP_MODE_LOITER:
                m_state.mode="LOITER";
                break;
              case CP_MODE_RTL:
                m_state.mode="RTL";
                break;
              case CP_MODE_CIRCLE:
                m_state.mode="CIRCLE";
                break;
              case CP_MODE_POSITION:
                m_state.mode="POSITION";
                break;
              case CP_MODE_LAND:
                m_state.mode="LAND";
                break;
              case CP_MODE_OF_LOITER:
                m_state.mode="OF_LOITER";
                break;
              case CP_MODE_DRIFT:
                m_state.mode="DRIFT";
                break;
              case CP_MODE_DUNE:
                m_state.mode="DUNE";
                break;
              case CP_MODE_SPORT:
                m_state.mode="SPORT";
                break;
              default:
                m_state.mode = "Unknown";
                break;

            }
          }
          else
          {
            switch(m_mode) // Plane.. not correct, but this is not Plane task!
            {
            default:
              m_external = true;
              m_state.mode = "Stabilize/unknown";
              break;
            case 3:
              trace("AUTO");
              m_state.mode = "Auto";
              m_external = false;
              break;
            case 5:
              trace("LOITER");
              m_state.mode = "Loiter";
              m_external = false;
              break;
            case 12:
              trace("DUNE");
              m_state.mode = "Dune";
              m_external = false;
              break;
            case 4:
              trace("GUIDED");
              m_state.mode = "Guided";
              m_external = false;
              break;
            }
          }
          dispatch(m_state);
        }

        void
        handleNavControllerPacket(const mavlink_message_t* msg)
        {
          if (!m_args.ardu_tracker)
            return;

          mavlink_nav_controller_output_t nav_out;
          mavlink_msg_nav_controller_output_decode(msg, &nav_out);
          trace("WP Dist: %d", nav_out.wp_dist);
          IMC::DesiredRoll d_roll;
          IMC::DesiredPitch d_pitch;
          IMC::DesiredHeading d_head;

          d_roll.value = Angles::radians(nav_out.nav_roll);
          d_pitch.value = Angles::radians(nav_out.nav_pitch);
          d_head.value = Angles::radians(nav_out.nav_bearing);

          if ((nav_out.wp_dist <= m_desired_radius + m_args.ltolerance)
             && (nav_out.wp_dist >= m_desired_radius - m_args.ltolerance)
             && (m_mode == CP_MODE_GUIDED))
          {
            if (!(m_cloops & IMC::CL_GUIDANCE)) {
              m_pcs.flags |= PathControlState::FL_LOITERING;
              trace("FL_LOITERING");
            }
          }

          // CHANGE
          if (m_args.wptmethod)
          {
			  if (!m_changing_wp
				 //&& (nav_out.wp_dist <= m_desired_radius + m_args.secs * m_gnd_speed)
				 //&& (nav_out.wp_dist >= m_desired_radius - m_args.secs * m_gnd_speed)
				 && (nav_out.wp_dist <= m_args.wptolerance)
				 && (m_mode == CP_MODE_GUIDED))
			  {
				if (!(m_cloops & IMC::CL_GUIDANCE)) {
				  m_pcs.flags |= PathControlState::FL_NEAR;
				  trace("FL_NEAR");
				}
			  }
          }
          else
          {
        	  if (!m_changing_wp
        	  				 && (nav_out.wp_dist <= m_desired_radius + m_args.secs * m_gnd_speed)
        	  				 && (nav_out.wp_dist >= m_desired_radius - m_args.secs * m_gnd_speed)
        	  				 && (m_mode == CP_MODE_GUIDED))
        	  			  {
        	  				if (!(m_cloops & IMC::CL_GUIDANCE)) {
        	  				  m_pcs.flags |= PathControlState::FL_NEAR;
        	  				  trace("FL_NEAR");
        	  				}
        	  			  }
          }

          dispatch(m_pcs);
          if (!(m_cloops & IMC::CL_ROLL)) {
            dispatch(d_roll);
          }

          if (!(m_cloops & IMC::CL_PITCH)) {
            dispatch(d_pitch);
          }

          if (!(m_cloops & IMC::CL_YAW)) {
            dispatch(d_head);
          }
        }

        void
        handleMissionItemPacket(const mavlink_message_t* msg)
        {
          mavlink_mission_item_t miss_item;
          mavlink_msg_mission_item_decode(msg, &miss_item);
          trace("Mission type: %d", miss_item.command);

          switch(miss_item.command)
          {
            default:
              m_critical = false;
              break;
            case MAV_CMD_NAV_TAKEOFF:
              m_critical = true;
              break;
            case MAV_CMD_NAV_LAND:
              m_critical = true;
              break;
          }
        }

        void
        handleHUDPacket(const mavlink_message_t* msg)
        {
          mavlink_vfr_hud_t vfr_hud;
          mavlink_msg_vfr_hud_decode(msg, &vfr_hud);

          IMC::IndicatedSpeed ias;
          IMC::TrueSpeed gs;


          ias.value = (fp64_t)vfr_hud.airspeed;
          gs.value = (fp64_t)vfr_hud.groundspeed;
          m_gnd_speed = (int)vfr_hud.groundspeed;

          dispatch(ias);
          dispatch(gs);
        }

        void
        handleSystemTimePacket(const mavlink_message_t* msg)
        {
          mavlink_system_time_t sys_time;
          mavlink_msg_system_time_decode(msg, &sys_time);

          time_t t = sys_time.time_unix_usec / 1000000;
          struct tm* utc;
          utc = gmtime(&t);

          m_fix.utc_time = utc->tm_hour * 3600 + utc->tm_min * 60 + utc->tm_sec + (sys_time.time_unix_usec % 1000000) * 1e-6;
          m_fix.utc_year = utc->tm_year + 1900;
          m_fix.utc_month = utc->tm_mon +1;
          m_fix.utc_day = utc->tm_mday;

          if (m_fix.utc_year>2014)
            m_fix.validity |= (IMC::GpsFix::GFV_VALID_TIME | IMC::GpsFix::GFV_VALID_DATE);

          dispatch(m_fix);
        }
      };
    }
  }
}

DUNE_TASK
