//***************************************************************************
// Copyright 2007-2014 Universidade do Porto - Faculdade de Engenharia      *
// Laboratório de Sistemas e Tecnologia Subaquática (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// https://www.lsts.pt/dune/licence.                                        *
//***************************************************************************
// Author: Eduardo Marques (original stage implementation)                  *
// Author: Pedro Calado                                                     *
// Author: Håvard L. Andersen (rewritten to new maneuver)					*
//***************************************************************************


#include <DUNE/Maneuvers/BarrierStages.hpp>
#include <DUNE/Coordinates/WGS84.hpp>
#include <DUNE/Math/Angles.hpp>

using namespace DUNE::Math;
using namespace DUNE::Coordinates;

namespace DUNE
{
	namespace Maneuvers
	{
		//! Value for y coordinate close to zero
	    //! Default constructor
	    BarrierStages::BarrierStages(const IMC::BarrierMan* maneuver, Tasks::Task* task)
	    {
	    	m_man = *maneuver;
	    	m_task = task;
	    	double x = 0;
	    	double y = 0;
	    	double d = m_man.length;

	    	m_stages.clear();

	    	for(int i = 0; i < m_man.numrounds; i++)
	    	{
	    		// point 1
	    		if (i == 0)
	    		{
	    			x = d/2;
					y = d/6;
					Stage s1("point1",x,y);
					m_stages.push_back(s1);
	    		}
	    		else
	    		{
	    			x = -d/2 + d/2;
					y = -(-d/6) + d/6;
					Stage s1("point1",x,y);
					m_stages.push_back(s1);
	    		}

	    		// point 2
				x = -d/2 + d/6;
				y = -d/6 + d/2;
				Stage s2("point2",x,y);
				m_stages.push_back(s2);

	    		// point 3
	    		x = -d/6 - d/2;
	    		y = -d/2 + d/2;
	    		Stage s3("point3",x,y);
	    		m_stages.push_back(s3);

	    		// point 4
	    		x = d/2 + d/2;
	    		y = -d/2 - d/2;
	    		Stage s4("point4",x,y);
	    		m_stages.push_back(s4);

	    		// point 5
	    		x = -d/2 - d/6;
				y = d/2 - d/2;
				Stage s5("point5",x,y);
				m_stages.push_back(s5);

				// point 6
				x = d/6 - d/2;
				y = d/2 - d/6;
				Stage s6("point6",x,y);
				m_stages.push_back(s6);

				// point 7
				x = d/2 - d/2;
				y = d/6 + d/6;
				Stage s7("point7",x,y);
				m_stages.push_back(s7);

				// point 8
				x = d/2 - d/6;
				y = -d/6 + d/2;
				Stage s8("point8",x,y);
				m_stages.push_back(s8);

				// point 9
				x = d/6 + d/2;
				y = -d/2 + d/2;
				Stage s9("point9",x,y);
				m_stages.push_back(s9);

				// point 10
				x = -d/2 - d/2;
				y = -d/2 - d/2;
				Stage s10("point10",x,y);
				m_stages.push_back(s10);

				// point 11
				x = d/2 + d/6;
				y = d/2 - d/2;
				Stage s11("point11",x,y);
				m_stages.push_back(s11);

				// point 12
				x = -d/6 + d/2;
				y = d/2 - d/6;
				Stage s12("point12",x,y);
				m_stages.push_back(s12);
	    	}

	    	m_curr = 0;
	    	m_sabs = Stage("undefined",0,0);
	    	m_index = 0;


	    }
	    bool
	        BarrierStages::getFirstPoint(double* lat, double* lon)
	        {
	          return getPoint(lat, lon);
	        }

	        bool
	        BarrierStages::getNextPoint(double* lat, double* lon)
	        {
	          return getPoint(lat, lon);
	        }

	        float
	        BarrierStages::getDistance(double* lat, double* lon)
	        {
	          return computeDistance(lat, lon);
	        }

	        float
	        BarrierStages::computeDistance(double* lat, double* lon)
	        {
	          BarrierStages::Stage last("", 0.0, 0.0);
	          BarrierStages::Stage zero("", 0.0, 0.0);

	          float total_distance = 0.0;
	          for(int i = 0; i < (int) m_stages.size(); i++)
	          {
	        	  total_distance = total_distance + m_stages[i].x + m_stages[i].y;
	        	  //total_distance = total_distance + std::sqrt(std::pow(m_stages[i-1].x -m_stages[i].x,2) + std::pow(m_stages[i-1].x -m_stages[i].y,2));
	          }


	          Angles::rotate(m_man.bearing, false, last.x, last.y);

	          *lat = m_man.lat;
	          *lon = m_man.lon;
	          Coordinates::WGS84::displace(last.x, last.y, lat, lon);

	          return total_distance;
	        }

	        bool
	        BarrierStages::getPoint(double* lat, double* lon)
	        {
	          const BarrierStages::Stage& new_stage = m_stages[m_index];
	          m_sabs.label = new_stage.label;
	          m_sabs.x += new_stage.x;
	          m_sabs.y += new_stage.y;


	          // Rotate according to row maneuver bearing angle
	          double dx = m_sabs.x, dy = m_sabs.y;

	          Angles::rotate(m_man.bearing, false, dx, dy);

	          if (m_task != NULL)
	            m_task->debug("%0.2f %0.2f -- %s", dx, dy, m_sabs.label);

	          // Calculate WGS-84 coordinates and fill DesiredPath message
	          *lat = m_man.lat;
	          *lon = m_man.lon;
	          Coordinates::WGS84::displace(dx, dy, lat, lon);

	          ++m_index;

	          if(m_index > m_stages.size())
	          {
	        	  return true;
	          }

	          return false;
	        }

	}
}
