//***************************************************************************
// Copyright 2007-2014 Universidade do Porto - Faculdade de Engenharia      *
// Laboratório de Sistemas e Tecnologia Subaquática (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// https://www.lsts.pt/dune/licence.                                        *
//***************************************************************************
// Author: Eduardo Marques (original stage implementation)                  *
// Author: Pedro Calado                                                     *
// Author: Håvard L. Andersen (rewritten to new maneuver)					*
//***************************************************************************


// DUNE headers.
#include <DUNE/Maneuvers/SectorStages.hpp>
#include <DUNE/Coordinates/WGS84.hpp>
#include <DUNE/Math/Angles.hpp>

using namespace DUNE::Math;
using namespace DUNE::Coordinates;

namespace DUNE
{
  namespace Maneuvers
  {
    //static const double pi = 3.14159265359;

    //! Default constructor.
    SectorStages::SectorStages(const IMC::SectorMan* maneuver, Tasks::Task* task)
    {
    	double pi = 3.14159265359;
    	m_man = *maneuver;
        m_task = task;

    	m_stages.clear();
    	int j_max = 3;

    	double x = 0;
    	double y = 0;
    	//char* label = "start";
    	int i = 1;
    	int j = 0;
    	// create a new Stage object
    	Stage start("start",x,y);
    	// Store the stage
    	m_stages.push_back(start);

    	double distance = m_man.length/2;

    	while(j < j_max)
    	{
    		switch(i)
    		{
    			case 1: x = distance * std::cos(pi/6 + j * pi/6);
    					y = distance * std::sin(pi/6 + j * pi/6);
    					break;
    			case 2: x = -distance * std::cos(pi/6 + j * pi/6) + distance * std::cos(pi/2 + j * pi/6);
    					y = -distance * std::sin(pi/6 + j * pi/6) + distance * std::sin(pi/2 + j * pi/6);
    					break;
    			case 3: x = -distance * std::cos(pi/2 + j * pi/6) + distance * std::cos(3*pi/2 + j * pi/6);
						y = -distance * std::sin(pi/2 + j * pi/6) + distance * std::sin(3*pi/2 + j * pi/6);
						break;
    			case 4: x = -distance * std::cos(3*pi/2 + j * pi/6) + distance * std::cos(11*pi/6 + j * pi/6);
						y = -distance * std::sin(3*pi/2 + j * pi/6) + distance * std::sin(11*pi/6 + j * pi/6);
						break;
    			case 5: x = -distance * std::cos(11*pi/6 + j * pi/6) + distance * std::cos(5*pi/6 + j * pi/6);
						y = -distance * std::sin(11*pi/6 + j * pi/6) + distance * std::sin(5*pi/6 + j * pi/6);
						break;
    			case 6: x = -distance * std::cos(5*pi/6 + j * pi/6) + distance * std::cos(7*pi/6 + j * pi/6);
						y = -distance * std::sin(5*pi/6 + j * pi/6) + distance * std::sin(7*pi/6 + j * pi/6);
						break;
    			case 7: x = -distance * std::cos(7*pi/6 + (j) * pi/6);
    					y = -distance * std::sin(7*pi/6 + (j) * pi/6);
    					j++;
    					i = 0;
    					break;
    		}
    		// create a new Stage object
    		Stage next("next",x,y);
    		// Store the stage
    		m_stages.push_back(next);

    		i++;

    	}
    	m_curr = 0;
    	m_sabs = Stage("undefined",0,0);
    	m_index = 0;

    }

    bool
    SectorStages::getFirstPoint(double* lat, double* lon)
    {
      return getPoint(lat, lon);
    }

    bool
    SectorStages::getNextPoint(double* lat, double* lon)
    {

      return getPoint(lat, lon);
    }

    float
    SectorStages::getDistance(double* lat, double* lon)
    {
      return computeDistance(lat, lon);
    }

    float
    SectorStages::computeDistance(double* lat, double* lon)
    {
      SectorStages::Stage last("", 0.0, 0.0);
      SectorStages::Stage zero("", 0.0, 0.0);

      float total_distance = 0.0;
      for(int i = 0; i < (int) m_stages.size(); i++)
      {
    	  total_distance = total_distance + m_stages[i].x + m_stages[i].y;
      	        	  //total_distance = total_distance + std::sqrt(std::pow(m_stages[i-1].x -m_stages[i].x,2) + std::pow(m_stages[i-1].x -m_stages[i].y,2));
      }

      Angles::rotate(m_man.bearing, false, last.x, last.y);

      *lat = m_man.lat;
      *lon = m_man.lon;
      Coordinates::WGS84::displace(last.x, last.y, lat, lon);

      return total_distance;
    }

    bool
    SectorStages::getPoint(double* lat, double* lon)
    {
      const SectorStages::Stage& new_stage = m_stages[m_index];
      m_sabs.label = new_stage.label;
      m_sabs.x += new_stage.x;
      m_sabs.y += new_stage.y;


      // Rotate according to row maneuver bearing angle
      double dx = m_sabs.x, dy = m_sabs.y;

      Angles::rotate(m_man.bearing, false, dx, dy);

      if (m_task != NULL)
        m_task->debug("%0.2f %0.2f -- %s", dx, dy, m_sabs.label);

      // Calculate WGS-84 coordinates and fill DesiredPath message
      *lat = m_man.lat;
      *lon = m_man.lon;
      Coordinates::WGS84::displace(dx, dy, lat, lon);

      ++m_index;
      if(m_index > m_stages.size())
      {
    	  return true;
      }

      return false;
    }
  }
}


