/*
 * Copyright (c) 2004-2014 Universidade do Porto - Faculdade de Engenharia
 * Laboratório de Sistemas e Tecnologia Subaquática (LSTS)
 * All rights reserved.
 * Rua Dr. Roberto Frias s/n, sala I203, 4200-465 Porto, Portugal
 *
 * This file is part of Neptus, Command and Control Framework.
 *
 * Commercial Licence Usage
 * Licencees holding valid commercial Neptus licences may use this file
 * in accordance with the commercial licence agreement provided with the
 * Software or, alternatively, in accordance with the terms contained in a
 * written agreement between you and Universidade do Porto. For licensing
 * terms, conditions, and further information contact lsts@fe.up.pt.
 *
 * European Union Public Licence - EUPL v.1.1 Usage
 * Alternatively, this file may be used under the terms of the EUPL,
 * Version 1.1 only (the "Licence"), appearing in the file LICENSE.md
 * included in the packaging of this file. You may not use this work
 * except in compliance with the Licence. Unless required by applicable
 * law or agreed to in writing, software distributed under the Licence is
 * distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF
 * ANY KIND, either express or implied. See the Licence for the specific
 * language governing permissions and limitations at
 * https://www.lsts.pt/neptus/licence.
 *
 * For more information please see <http://lsts.fe.up.pt/neptus>.
 *
 * Author: havard
 * Mar 24, 2014
 */
package pt.lsts.neptus.mp.maneuvers;

import java.util.Vector;

/**
 * @author havard
 * Functions for computing the waypoints for each search pattern.
 *
 */
public class SARManeuversUtil {
    protected static final int X = 0, Y = 1, Z = 2, T = 3;
    
    private SARManeuversUtil(){
        
    }
    
    public static Vector<double[]> calcSquareSearchPoints(double distance, double S,
             double bearingRad) {
        
        Vector<double[]> newPoints = new Vector<double[]>();
        int i = 1;
        int a = 1;
        double[] point = {0,0,0,-1};
        double px = 0;
        double py = 0;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        boolean morePoints = true;
        
        while (morePoints)
        {
            switch(i)
            {
                case 1: px = px + a * S;
                        break;
                case 2: py = py + a * S;
                        a++;
                        break;
                case 3: px = px - a * S;
                        break;
                case 4: py = py - a * S;
                        a++;
                        i = 0;
                        break;
            }
            i++;
            point = new double[]{px,py,0,-1};
            
            
            if (a * S > distance + S)
                morePoints = false;
            else
                newPoints.add(point);
                
        }
        // rotate the pattern if necessary
        if (bearingRad != 0)
        {
            Vector<double[]> rotatedPoints = new Vector<double[]>();
            double[] rotpoint = {0,0,0,-1};
            for (i = 0; i <newPoints.size(); i++)
            {
                double r1 = Math.cos(bearingRad) * newPoints.get(i)[0] - Math.sin(bearingRad) * newPoints.get(i)[1];
                double r2 = Math.sin(bearingRad) * newPoints.get(i)[0] + Math.cos(bearingRad) * newPoints.get(i)[1];
                rotpoint = new double[]{r1,r2,0,-1};
                rotatedPoints.add(rotpoint);
            }
            return rotatedPoints;
        }
        return newPoints; 
    }
    public static Vector<double[]> calcSectorSearchPoints(double distance,
            double bearingRad) {
       
       Vector<double[]> newPoints = new Vector<double[]>();
       int i = 1;
       int j = 0;
       double px = 0;
       double py = 0;
       double[] point = {0,0,0,-1};
       newPoints.add(point);
       while (j <3)
       {
           
           switch(i)
           {
               case 1: px = (distance/2) * Math.cos(Math.toRadians(30 + j * 30));
                       py = (distance/2) * Math.sin(Math.toRadians(30 + j * 30));
                       break;
               case 2: px = (distance/2) * Math.cos(Math.toRadians(90 + j * 30));
                       py = (distance/2) * Math.sin(Math.toRadians(90 + j * 30));
                       break;
               case 3: px = (distance/2) * Math.cos(Math.toRadians(270 + j * 30));
                       py = (distance/2) * Math.sin(Math.toRadians(270 + j * 30));
                       break;
               case 4: px = (distance/2) * Math.cos(Math.toRadians(330 + j * 30));
                       py = (distance/2) * Math.sin(Math.toRadians(330 + j * 30));
                       break;
               case 5: px = (distance/2) * Math.cos(Math.toRadians(150 + j * 30));
                       py = (distance/2) * Math.sin(Math.toRadians(150 + j * 30));
                       break;
               case 6: px = (distance/2) * Math.cos(Math.toRadians(210 + j * 30));
                       py = (distance/2) * Math.sin(Math.toRadians(210 + j * 30));
                       break;
               case 7: px = 0;  // go center
                       py = 0;
                       j++;
                       i = 0;
                       break;
           }
           point = new double[]{px,py,0,-1};
           newPoints.add(point);
           
           i++;

       }
       // rotate the pattern if necessary
       if (bearingRad != 0)
       {
           Vector<double[]> rotatedPoints = new Vector<double[]>();
           double[] rotpoint = {0,0,0,-1};
           for (i = 0; i <newPoints.size(); i++)
           {
               double rpx = Math.cos(bearingRad) * newPoints.get(i)[0] - Math.sin(bearingRad) * newPoints.get(i)[1];
               double rpy = Math.sin(bearingRad) * newPoints.get(i)[0] + Math.cos(bearingRad) * newPoints.get(i)[1];
               rotpoint = new double[]{rpx,rpy,0,-1};
               rotatedPoints.add(rotpoint);
           }
           return rotatedPoints;
       }
       return newPoints; 
   }
    /**
     * @param length
     * @param numRounds
     * @param bearingRad
     * @return
     */
    public static Vector<double[]> calcBarrierPatrolPoints(double distance, int numRounds, double bearingRad) {
     // rotate the pattern if necessary
        Vector<double[]> newPoints = new Vector<double[]>();
        double[] point = {0,0,0,-1};
        double px = 0;
        double py = 0;
        
        // point 1
        px = distance/2;
        py = distance/6;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 2
        px = distance/6;
        py = distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 3
        px = -distance/2;
        py = distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 4
        px = distance/2;
        py = -distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        
        // point 5
        px = -distance/6;
        py = -distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 6
        px = -distance/2;
        py = -distance/6;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 7
        px = -distance/2;
        py = distance/6;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        
        // point 8
        px = -distance/6;
        py = distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 9
        px = distance/2;
        py = distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 10
        px = -distance/2;
        py = -distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        

        // point 11
        px = distance/6;
        py = -distance/2;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        
        // point 12
        px = distance/2;
        py = -distance/6;
        point = new double[]{px,py,0,-1};
        newPoints.add(point);
        
        
        if (bearingRad != 0)
        {
            Vector<double[]> rotatedPoints = new Vector<double[]>();
            double[] rotpoint = {0,0,0,-1};
            for (int i = 0; i <newPoints.size(); i++)
            {
                double rpx = Math.cos(bearingRad) * newPoints.get(i)[0] - Math.sin(bearingRad) * newPoints.get(i)[1];
                double rpy = Math.sin(bearingRad) * newPoints.get(i)[0] + Math.cos(bearingRad) * newPoints.get(i)[1];
                rotpoint = new double[]{rpx,rpy,0,-1};
                rotatedPoints.add(rotpoint);
            }
            return rotatedPoints;
        }
        return newPoints;
    }

    
    public static void main(String[] args) {
       /* Vector<double[]> points = SARManeuversUtil.calcSquareSearchPoints(200, 20, 0);
        
        System.out.println(points.size());
        for(int i = 0; i < points.size(); i++)
        {
            System.out.println("i: " + i + " px " + points.get(i)[0] + " py " + points.get(i)[1] );
            //double[] h = points.get(i);
            //System.out.println(""+ h[0] + " " + h[1] +"  " + h[2]);
        }
        //double[] hh = {2,3,4,5};
        //hh[1] = 87;
        //System.out.println(hh[1]);*/
    }

    
    
}
